package asp;

import java.io.IOException;
import java.io.StreamTokenizer;
import java.util.ArrayList;
import java.util.Iterator;

/**
 * 
 * @author Alexander Häusler
 * @author Marco Müller
 *
 */
public class NonTerminal extends Term {
	
	private String name;
	private ArrayList<Term> paras = new ArrayList<Term>();

	
	NonTerminal(String[] str) {
		super(false);
		this.name = str[0];
		
		for (int i=1; i<str.length; i++)
			this.paras.add(new Terminal(str[i]));
	}
	
	
	NonTerminal(String name) {
		super(false);
		this.name = name;
	}
	
	
	/**
	 * Appends the given term to the list of parameters. Parameters can be terminal or
	 * nonterminal symbols.
	 * 
	 * @param term
	 */
	protected void addParameter(Term term) {
		this.paras.add(term);
	}
	
	
	/**
	 * Appends the given terms to the list of parameters. Parameters can be terminal or
	 * nonterminal symbols.
	 * 
	 * @param terms
	 */
	protected void addParameters(Term[] terms) {
		for (Term term : terms)
			this.paras.add(term);
	}
	
	
	/**
	 * @return The GDL representation of the object
	 */
	public String toGDLString() {
		String buffer;
		buffer = "(" + this.name.trim();
		
		for (Term term : this.paras)
			buffer += " " + term.toGDLString();
		
		return buffer.trim() + ")";
	}

	
	/**
	 * @return The ASP representation of the object
	 */
	public String toASPString() {
		String buffer;
		buffer = this.name + "(";
		
		if (!this.paras.isEmpty()) {			
			Iterator<Term> iter = this.paras.iterator();
			
			buffer += iter.next().toASPString();
			
			while (iter.hasNext())
				buffer += "," + iter.next().toASPString();		
		}
			
		return buffer + ")";
	}

	
	/**
	 * The NonTerminal object is built step by step by the stream of a StreamTokenizer.<br>
	 * <u>Attention:</u> The leading '(' must already be removed from the Tokenizer
	 *
	 * @param tokenizer
	 * @throws IOException
	 */
	public static NonTerminal parseGDLString(StreamTokenizer tokenizer) throws IOException {
		NonTerminal nt;
		
		tokenizer.nextToken();
		nt = new NonTerminal(tokenizer.sval);
		
		loop:
		while (tokenizer.nextToken() != StreamTokenizer.TT_EOF) {
			switch (tokenizer.ttype) {
				case StreamTokenizer.TT_NUMBER:
					nt.addParameter(new Terminal(String.valueOf(((int) tokenizer.nval))));
					break;
				case StreamTokenizer.TT_WORD:
					nt.addParameter(new Terminal(tokenizer.sval));
					break;
				case '(':
					nt.addParameter(NonTerminal.parseGDLString(tokenizer));
					break;
				case ')':
					break loop;
			}
			
		}
		
		return nt;
	}

	
	/**
	 * @param name A NonTerminal object consists of a predicate name and some parameters
	 * @param tokenizer
	 * @return The parsed NonTerminal object
	 * @throws IOException
	 */
	public static NonTerminal parseASPString(String name, StreamTokenizer tokenizer) throws IOException {
		NonTerminal nt = new NonTerminal(name);
		String buffer;
		
		loop:
		while (tokenizer.nextToken() != StreamTokenizer.TT_EOF) {
			switch (tokenizer.ttype) {
				case StreamTokenizer.TT_NUMBER:
					nt.addParameter(new Terminal(String.valueOf(((int) tokenizer.nval))));
					break;
					
				case StreamTokenizer.TT_WORD:
					buffer = tokenizer.sval;
					if (tokenizer.nextToken() == '(') {
						//tokenizer.nextToken();
						nt.addParameter(NonTerminal.parseASPString(buffer, tokenizer));
					}
					else if (tokenizer.ttype == ')') {
						nt.addParameter(new Terminal(buffer));
						break loop;
					}
					else if (tokenizer.ttype == ',') {
						nt.addParameter(new Terminal(buffer));
					}
					break;
					
				case ')':
					break loop;
			}
			
		}
		
		return nt;
	}
	
	

}
