package asp;

import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * 
 * @author Alexander Häusler
 * @author Marco Müller
 * 
 * @version 1.0
 * 
 */
public class XASPSolverController extends Thread {
	
	private boolean solved = false;
	
	// a reference to the ASPController is needed as long as the '--wait=no' option does not work
	private XASPController aspController;
	
	// oClingo and controller.py processes
	private ProcessBuilder  pb, ob;
	private Process controller, oclingo;
	
	// steps found by the oClingo process
	private ArrayList<Step> steps;
	
	// steps given to oClingo by the controller
	private ArrayList<Step> externalSteps;
	
	// parsed oClingo file
	private File encoding = null;
	private String filename;
	
	// controller.py's stdin and stdout
	//private DataOutputStream stdin;
	private PrintStream stdin;
	private BufferedReader stdout;
	
	// oClingo stdout
	private BufferedReader oClingoOut;
	
	// RegEx pattern for answer set parsing
	private static final String DOES_PATTERN = "does\\(\\w+,(.*),\\d+\\)";
	
	// our Role
	private String role;
	
	

	
	public XASPSolverController(XASPController aspController, String role, String filename) {
		this.aspController = aspController;
		this.filename	   = filename;
		this.role 		   = role;
		
		// same references
		this.steps 		   = aspController.steps;
		this.externalSteps = aspController.externalSteps;
	}

	
	public void putInStep(Step s) {
		externalSteps.add(s);
		
		/* While the '--wait=no' option is not working the entire solving process must be
		 * restarted manually. Otherwise a reference to the ASPContoller would not be needed. */
		this.solved = false;
		this.aspController.restart();
	}
	
	
	public boolean isSolved() {
		return this.solved;
	}
	
	
	/**
	 * First the method tries to terminate the processes the normal way.
	 * But if oClingo is still in process of solving, the normal way has no effect 
	 * or will take too much time. So both processes are killed additionally to avoid the
	 * waste of resources, which can easily be extreme!
	 */
	public void kill() {
		try {
			stdin.print("#stop.");
			stdin.flush();
		} catch (Exception e) {
			//e.printStackTrace();
		}
		
		
		this.steps.clear();
		this.solved = false;
		
		// force termination
		if (controller != null)
			controller.destroy();
		
		if (oclingo != null)
			oclingo.destroy();
	}
	
	
	/**
	 * The method is called to provide the oClingo process all steps which 
	 * have already been done.
	 */
	private void putInExternalSteps() {
		if (this.externalSteps.isEmpty())
			return;

		String tmp;
		int counter = 0;
		
		try {
			stdin.print("#step 1.\n");
			stdin.flush();
			
			for (Step s: this.externalSteps) {
				tmp = s.toASPString();
				System.out.println(tmp);
				tmp = "extDoes(" + role + "," + tmp.substring(0,tmp.length() - 1) + "," + (++counter) + ").\n";
				
				stdin.print(tmp);
				stdin.flush();
			}
			
			stdin.print("#endstep.\n");
			stdin.flush();
		} catch(Exception e) {}
	}
	
	
	private void readAnswerSet() {
		String  buffer;
		Matcher matcher;
		Pattern doesRegEx = Pattern.compile(DOES_PATTERN);
		
		try {
			/* Answer sets are read as long as the inputStream is alive.
			 * If the stream is closed an exception is thrown while reading from the stream
			 * and the infinite loop is left */
			buffer = stdout.readLine();
			while (!buffer.startsWith("Answer")) 
				buffer = stdout.readLine();
			this.solved = false;
			
			// empty current step list
			this.steps.clear();
			
			// a single answer set is a list of does/3 predicates
			// does/3 is parsed and stored
			matcher = doesRegEx.matcher(stdout.readLine());
			while (matcher.find()) {
				this.steps.add(Step.parseASPString(matcher.group(1)));
				matcher = doesRegEx.matcher(stdout.readLine());
			}
			
			this.solved = true;
			this.aspController.stepsIterator = this.steps.iterator();
	
			// test output
			System.out.println("\nANSWER:");
			for (Step s: this.steps)
				System.out.println(s.toGDLString());
			
			stdin.print("#stop.");
			
		} catch(NullPointerException e) {
			
		} catch (Exception e) {
		} 
	}
	

	public void run() {
		System.out.println("Solver start");
		this.encoding = new File(this.filename);
		
		if (!this.encoding.isFile()) {
			System.err.println("file location does not exist");
			return;
		}
	
		/* each time the solverController is started or restarted the list of found steps
		 * is cleared to avoid the unification of two answer sets to a single one. */
		this.steps.clear();
	
		ob = new ProcessBuilder("./asp/bin/oclingo", this.encoding.getPath());
		
		try {
			oclingo = ob.start();
			System.out.println("oclingo started!");	
			oClingoOut = new BufferedReader(new InputStreamReader(oclingo.getErrorStream()));
			
			String buffer;
			String[] firstLine = null;
			
			
			/* the oClingo process must be ready ('Waiting for connection...') 
			 * before the controller can be started */
			while ((buffer = oClingoOut.readLine()) != null) {
				//System.out.println("Lese Waiting");
				firstLine = buffer.split(" ");
				
				if (firstLine[0].equalsIgnoreCase("Waiting")) {
					//System.out.println("Port: " + firstLine[5]);
					break;
				}
			}

			
			/* 'python -u' force stdin, stdout and stderr to be totally unbuffered (man python).
			 * This options is needed to get immediately any output sent by the controller.
			 * 
			 * Currently the '--wait' option is not supported! */
			pb = new ProcessBuilder("python", "-u", "./asp/bin/controller.py", "--wait=no", "--port=" + firstLine[5]);
			
			controller = pb.start();
			stdin 	   = new PrintStream(controller.getOutputStream(), true, "US-ASCII");
			stdout 	   = new BufferedReader(new InputStreamReader(controller.getInputStream()));
			System.out.println("controller started");			


			if (!externalSteps.isEmpty())
				putInExternalSteps();
			else { 
				stdin.println("#endstep.");
				stdin.flush();
			}
			
			System.out.println("read AS");			
			this.readAnswerSet();

			oClingoOut.close();
			stdin.close();
			stdout.close();
			
			System.out.println("Total number of found steps: " + this.steps.size());
			
			// force termination
			oclingo.destroy();
			controller.destroy();
			
		} catch (IOException e) {
		}
		
		System.out.println("~~ ASP thread terminates now!");
	}

}
