package basic;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintStream;
import java.io.StringReader;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.regex.MatchResult;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import asp.ASPException;
import asp.XASPController;
import basic.ai.AI;
import basic.ai.SimpleMC;
import basic.ai.mcts.MCTS;
import basic.logic.GameLogic;
import basic.logic.Move;
import basic.logic.PrologLogic;
import basic.logic.Role;
import basic.logic.State;
import basic.prepro.Preprocessor;
import basic.prepro.STCompoundTermTranslator;
import basic.prepro.STListStatesTranslator;
import basic.prepro.STTranslator;
import basic.prepro.Selector;

/**
 * GGP Basicplayer of the University of Potsdam
 * 
 * Original part of Centurio:
 * Centurio a General Game Player
 *
 * Copyright (C) 2008 Felix Maximilian Möller, Marius Schneider and Martin Wegner
 *
 * This file is part of Centurio.
 *
 * Centurio is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
 *
 * Centurio is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with Centurio. If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * This class is responsible for starting a game, sending messages to the gamemaster and
 * receiving them. The class also controls the end of a game and the behavior of the slaves.
 *
 * @author Felix Maximilian Möller
 * @author Marius Schneider
 * @author Martin Wegner
 * @author University Dresden
 * @version 1.0
 */
public class GamePlayer extends AbstractGamePlayer {

	private String matchID;
	private Role ourRole; // the place the Centurios move stands in the list of moves of all Players
	private Short startClock, playClock;
	private Boolean firstMove;				// needed -> avoids putting moves of the first play message into translatedMove
	private STCompoundTermTranslator moveTranslator = new STCompoundTermTranslator();
	private Collection translatedMove = new ArrayList(); // contains the translated moves of the play/stop message
	private GameLogic gl;					// capsules all game description reasoning (-> prolog)
	private AI slave;						// instance of KI (rating of moves)
	private State currentState;
	private Move executedMove;
	
	private XASPController asp;
	
	//gameproperties
	private Boolean soloGame;				// game with only one role
	private static GamePlayer gp;			
	private int moveCounter;				// counts the number of moves which were played
	// how many processors are available to VM
	private int NumberOfProcessors = 1; // only set higher if your Player is ready for multi-threading -> Runtime.getRuntime().availableProcessors();
	
	/**
	 * Constructor
	 * @param port
	 * @throws IOException
	 */
	public GamePlayer(int port) throws IOException {
		super(port);
	}

	/**
	 * saves the gameplayer instance
	 * @param object
	 */
	static public void saveGamePlayerObject(GamePlayer object) {
		gp = object;
	}

	/**
	 * wakes up the game player
	 */
	static public void wakeUp() {

		synchronized(gp) {
			gp.notify();
		}

	}

	/**
	 * returns the boolean value true if the game is a solo game
	 * @return the boolean value true if the game is a solo game
	 */
	public Boolean getSoloGame() {
		return soloGame;
	}

	/**
	 * set the boolean to identify if the game is a solo game
	 * @param soloGame
	 */

	public void setSoloGame(Boolean soloGame) {
		this.soloGame = soloGame;
	}
	
	/**
	 * getter method for translated move
	 * @return
	 */
	public Collection getTranslatedMove() {
		return translatedMove;
	}

	/**
	 * this method is called when a new match begins
	 */
	protected void commandStart(String msg){
		// msg="(START <MATCH ID> <ROLE> <GAME DESCRIPTION> <STARTCLOCK> <PLAYCLOCK>)
		// e.g. msg="(START tictactoe1 white ((role white) (role black) ...) 1800 120)" means:
		//       - the current match is called "match0815"
		//       - your role is "white",
		//       - after at most 1800 seconds, you have to return from the commandStart method
		//       - for each move you have 120 seconds
		
		Long startTime = System.currentTimeMillis();
		moveCounter = 0;

		Thread.currentThread().setPriority(8); // GamePlayer-thread gets higher priority as slave threads
		
		// Start Nachricht wird per regul�ren Ausdruck zerlegt
		Pattern regex = Pattern.compile("^\\(START\\s+(\\S+)\\s+(\\S+)\\s+\\((.+)\\)\\s+(\\d+)\\s+(\\d+)\\)$", Pattern.DOTALL);
		Matcher matcher = regex.matcher(msg);

		if(matcher.matches()) {

			// filtern von Information aus der Startmessage
			matchID = matcher.group(1);
			ourRole = new Role(matcher.group(2).toLowerCase());
			String gameDescription = matcher.group(3);
			startClock = Short.parseShort(matcher.group(4));
			playClock = Short.parseShort(matcher.group(5));

			try {
				// anlegen der kif-Datei
				FileWriter fw = new FileWriter("./kif/" + matchID + ".kif");
				fw.write(gameDescription);
				fw.close();

			}

			catch(IOException e) {
				System.err.println("(GamePlayer - commandStart) error message: write gameDescription - IOException\n\n" + e.getMessage() + "\n\n" + e.toString());				
			}

			System.out.println("Basicplayer's role is: " + ourRole.get());

			firstMove = true; // avoids putting moves of the first play message into translatedMove

			try { // parses the game description

				STListStatesTranslator translator = new STListStatesTranslator();
				translator.translate(new StringReader(gameDescription), new PrintStream(new File("./pl/" + matchID + ".pl")));

			}

			catch(IOException e) {
				System.err.println("(GamePlayer - commandStart) error message: gameDescription translator - IOException\n\n" + e.getMessage() + "\n\n" + e.toString());
			}

			catch(STListStatesTranslator.ParseException e) {
				System.err.println("(GamePlayer - commandStart) error message: gameDescription translator - STListStatesTranslator.ParseException\n\n" + e.getMessage() + "\n\n" + e.toString());
			}

			// delete all slaves and eclipse prolog instances
			if (slave != null)
				((Thread) slave).interrupt();
			slave = null;
			
			/*
			 * TODO Exception verhindert commandStart beim 2. Durchlauf
			 */
			if(gl != null) {
				gl.destroy();
				gl = null;
			}

			asp = new XASPController(matchID,(String) ourRole.get());
			try {
				this.soloGame = asp.isSoloGame();
			} catch(ASPException e) {
				this.soloGame = false;
			}
			
			System.out.println("###########" + this.soloGame);
			
			/*
			if (this.soloGame)
				asp.start();
			*/
			
			
			gl = new PrologLogic(matchID);  //connection to the (prolog) engine for the game logic
			
			if(!this.soloGame) 
				slave = new SimpleMC(new PrologLogic(matchID),ourRole);
			else slave = new MCTS(new PrologLogic(matchID), ourRole);
			// find gameproperties	
			currentState = gl.initState();
			slave.initState(currentState);
			
			((Thread) slave).setPriority(5);
			((Thread) slave).start();
				
			Preprocessor prep = new Selector();
			prep.select(gl);

            synchronized (gp) { // lets the slaves working until the start clock is over minus 1 second overhead for communication

            	try {
            		gp.wait((startClock - 1) * 1000 - (System.currentTimeMillis() - startTime));
            	}

            	catch(InterruptedException e) {
            		System.err.println("(GamePlayer - commandStart) error message: Sleeping - InterruptedException\n\n" + e.getMessage() + "\n\n" + e.toString());
            	}

            }
		}
		else
			System.err.println("Could not interprete START message from server.");
		// "READY"-Nachricht wird automatisch nun gesendet
	}

	/**
	 * this method is called once for each move
	 * @return the move of this player
	 */
	protected String commandPlay(String msg){
		// msg="(PLAY <MATCH ID> <JOINT MOVE>)
		// <JOINT MOVE> will be NIL for the first PLAY message and the list of the moves of all the players in the previous state
		// e.g. msg="(PLAY tictactoe1 NIL)" for the first PLAY message
		//   or msg="(PLAY tictactoe1 ((MARK 1 2) NOOP))" if white marked cell (1,2) and black did a "noop".
		
		// Done:
		//    - calculate the new state from the old one and the <JOINT MOVE>
		//    - use the time to find the best of your possible moves in the current state
		//    - return your move (instead of "NIL") on time (before PLAYCLOCK is over!)
		System.out.println("CALL commandPlay()");
		Long startTime = System.currentTimeMillis();
		moveCounter++;

		Pattern regex = Pattern.compile("^\\(PLAY\\s+(\\S+)\\s+\\(?(.+)\\)?\\)$", Pattern.DOTALL);
		Matcher matcher = regex.matcher(msg);

		if(matcher.matches()) {

			if(matchID.equals(matcher.group(1))) {

				/*
				 * ###############################################################################
				 *  TODO Exceptions behindert lesen der Antwortmenge von oClingo ab dem 2 Schritt
				 * ###############################################################################
				 */
				if(!firstMove) { // it is not the first play message and the moves of the play message will be translated and stored into translatedMove

					String moves = matcher.group(2);
					
					try {
						translatedMove = moveTranslator.translate(new StringReader(moves), false);
					}
					catch(IOException e) {
						System.err.println("(GamePlayer - commandPlay) error message: move translator - IOException\n\n" + e.getMessage() + "\n\n" + e.toString());
					}
					catch(STCompoundTermTranslator.ParseException e) {
						System.err.println("(GamePlayer - commandPlay) error message: move translator - STCompoundTermTranslator.ParseException\n\n" + e.getMessage() + "\n\n" + e.toString());
					}
					
					

					
					if (!soloGame) {
						try {
							executedMove = new Move(translatedMove);
							currentState = gl.nextState(currentState, executedMove);
							slave.setNewState(executedMove);
//							currentState = gl.nextState(currentState, new Move(jpl.Util.textToTerm(new basic.logic.k2p().translate(moves))));
//							slave.setNewState(new Move(jpl.Util.textToTerm(new basic.logic.k2p().translate(moves))));
						} catch (Exception e) {
							List<Move> MoveList = gl.legal(currentState);
							Collections.shuffle(MoveList);
							return STTranslator.printKIF(MoveList.get(0).get());
						}
						
					}
					else{
						executedMove = new Move(translatedMove);
						currentState = gl.nextState(currentState, executedMove);
						//slave.initState(currentState);
						slave.setNewState(executedMove);
					}
					
					
				}

				else {
					firstMove = false; // detects the first play message

				}

				Object nextMove = "NIL";

				synchronized (gp) { // lets the slaves working until the play clock is over minus 1 second overhead for communication and 0,1 seconds for getBestMove

		           	try {
	            		gp.wait((playClock - 1) * 1000 - 100 - (System.currentTimeMillis() - startTime));
	            	}

	            	catch(InterruptedException e) {
	            		System.err.println("(GamePlayer - commandPlay) error message: Sleeping - InterruptedException\n\n" + e.getMessage() + "\n\n" + e.toString());
	            	}

	            }

				/*
				 * TODO oClingo oder MCTS?
				 */
				/*
				if (this.soloGame && asp.isSolved()) {
					try {
						System.out.println("ASP answers!");
						return asp.getNextStep();
					} catch(ASPException e) {
						System.out.println("ASP Exception");
						e.printStackTrace();
					}
				}
				*/
				// call getBestMove of KI Class 
				try {
					nextMove = slave.getBestMove(ourRole);
				} catch (Exception e) {
					nextMove = gl.legal(currentState).get(0).get();
				}
				String nextMoveString = STTranslator.printKIF(nextMove);
//				String nextMoveString = new basic.logic.k2p().translate2kif(nextMove.toString());
				/*
				if (this.soloGame) {
					System.out.println("putInStep");
					asp.putInStep(nextMoveString);
				}
				*/
				return STTranslator.printKIF(nextMove);
//				return new basic.logic.k2p().translate2kif(nextMove.toString());
			}

			else {

				System.err.println("Wrong match ID in PLAY message.");

				return "NIL";

			}

		}

		else {

			System.err.println("Could not interprete PLAY message from server.");

			return "NIL";

		}

	}

	/**
	 * this method is called if the match is over
	 */

	protected void commandStop(String msg){
		// msg="(STOP <MATCH ID> <JOINT MOVE>)
		
		// Done:
		//    - clean up the GamePlayer for the next match
		//    - be happy if you have won, think about what went wrong if you have lost ;-)

		Pattern regex = Pattern.compile("^\\(STOP\\s+(\\S+)\\s+\\(?(.+)\\)?\\)$", Pattern.DOTALL);
		Matcher matcher = regex.matcher(msg);

		if(matcher.matches()) {

			if(matchID.equals(matcher.group(1))) {

				String moves = matcher.group(2);
				
				/*
				if (this.soloGame)
				asp.kill();
*/

				try { // the moves of the stop message will be translated and stored into translatedMove
					translatedMove = moveTranslator.translate(new StringReader(moves), false);
				}

				catch(IOException e) {
					System.err.println("(GamePlayer - commandStop) error message: move translator - IOException\n\n" + e.getMessage() + "\n\n" + e.toString());
				}

				catch(STCompoundTermTranslator.ParseException e) {
					System.err.println("(GamePlayer - commandStop) error message: move translator - STCompoundTermTranslator.ParseException\n\n" + e.getMessage() + "\n\n" + e.toString());
				}
				

				if (!soloGame) {
					executedMove = new Move(translatedMove);
					currentState = gl.nextState(currentState, executedMove);
					try {
						slave.setNewState(executedMove);
					} catch (Exception e) {
						// TODO: handle exception
					}
				}
				else {
					executedMove = new Move(translatedMove);
					currentState = gl.nextState(currentState, new Move(translatedMove));
					try {
						slave.setNewState(executedMove);
					} catch (Exception e) {
						// TODO: handle exception
					}
					//slave.initState(currentState);
				}
					
//					currentState = gl.nextState(currentState, new Move(jpl.Util.textToTerm(new basic.logic.k2p().translate(moves))));
//					slave.setNewState(new Move(jpl.Util.textToTerm(new basic.logic.k2p().translate(moves))));
				//}
				
				
				int[] scores = gl.goal(currentState);
		
				for(int i = 0; i < scores.length; i++) {
					System.out.println("Player " + STTranslator.printKIF(gl.getRoles().get(i).get()) + " has " +  scores[i] + " points."); // print the scores of all players
				}
				
				System.out.println("Basicplayer's role was: " + ourRole.get());



				// delete all slaves and eclipse prolog instances

				((Thread) slave).interrupt();
				//clean up gameplayer
				slave = null;
				gl.destroy();
				gl = null;
				System.gc();

			}

			else
				System.err.println("Wrong match ID in STOP message.");

		}

		else
			System.err.println("Could not interprete STOP message from server.");

	}

}
