package basic.ai.mcts;

import basic.ai.Node;
import basic.logic.GameLogic;
import basic.logic.Move;
import basic.logic.State;

public class UCTSelectionPolicy implements TreePolicy {

	private static final double CONST_P = 50;
	
	private GameLogic gl;
	
	public UCTSelectionPolicy(GameLogic gameLogic) {
		this.gl = gameLogic;
	}
	
	@Override
	public Node selectNode(Node node) {
		while(!node.isTerminal()) {
			
			if (node.isFullyExpanded()) {
				node = selectBest(node);
			} else {
				return expand(node);
			}
		}
		return node;
	}
	
	public Node selectBest(Node node) {
		double bestScore = Double.NEGATIVE_INFINITY;
		Node bestNode = null;
		
		for (Node child : node.getChildren()) {
			double curUCTScore = calcUCTScore(child, node);
			if (curUCTScore > bestScore) {
				bestScore = curUCTScore;
				bestNode = child;
			}
		}
		
		return bestNode;
	}
	
	public Node expand(Node node) {
		Move nextMove = node.getUnextendedMoves().iterator().next();
		State nodesState = gl.nextState(node.getState(), nextMove);
		boolean isTerminal = gl.isTerminal(nodesState);
		Node newNode = node.addChild(
				nextMove, 
				nodesState, 
				isTerminal ? null : gl.legal(nodesState),
				isTerminal
			);
		
		return newNode;
	}
	
	public double calcUCTScore(Node node, Node parent) {
		return 
			(((double) node.getRewardSum())/((double) node.getVisits())) + CONST_P * Math.sqrt((2.0*Math.log(parent.getVisits()))/((double) node.getVisits()));
	}
//((node.total_nodes() - node.getVisits())/(double) node.total_nodes())
}
